==> healthrian_ble_support_for_wearecg12-1.0.17.tar.gz: lib/src/mixin/ecg.dart <==
// ignore_for_file: public_member_api_docs

import 'dart:developer';

import 'package:collection/collection.dart';
import 'package:fl_chart/fl_chart.dart';
import 'package:healthrian_ble_support_for_wearecg12/'
    'healthrian_ble_support_for_wearecg12.dart';
import 'package:healthrian_common_support/healthrian_common_support.dart';
import 'package:lzma/lzma.dart';

mixin Ecg {
  /// [maxBufferStreamEcg] decides a maximum length of buffered Stream.
  ///
  /// For 40 seconds buffer and 250 sampling rate,
  /// It uses 40 * 250 by default.
  int maxBufferStreamEcg = 40 * 250;

  /// [lengthBufferStreamEcg] decides a length of buffered Stream,
  /// after flushing buffer when reaching a maximum size.
  ///
  /// For 30 seconds buffer and 250 sampling rate,
  /// It uses 30 * 250 by default.
  int lengthBufferStreamEcg = 30 * 250;

  /// [screenBufferStreamingEcg] decides a length of buffered Stream,
  /// which is shown to screen.
  ///
  /// For 5 seconds buffer and 250 sampling rate,
  /// It uses 5 * 250 by default.
  int screenBufferStreamingEcg = 5 * 250;

  /// [refreshBufferEcg] decides a length to be hidden after refreshed.
  ///
  /// For 0.5 seconds range to be hidden and 250 sampling rate,
  /// It use 250 ~/ 2 by default.
  int refreshBufferEcg = 250 ~/ 2;

  /// [isFilteredEcg] decides whether ecg signal is filtered.
  bool isFilteredEcg = true;

  /// [isCompressedEcg] decides whether ecg signal is compressed.
  ///
  /// If it is `true`, ecg signal is decompressed by LZMA algorithm.
  bool isCompressedEcg = false;

  final _dataFilter = DataFilter(8);

  final _adaptiveHR = AdaptiveHR(0);

  bool _isPaused = false;

  bool _isMeasuring = false;

  /// [skipSampleCount] is the number of samples to be skipped.
  /// For example, [skipSampleCount] is 5 by default.
  /// The plotting stream only emits a 1/5 samples,
  /// which discards 4/5 samples.
  int skipSampleCount = 5;

  /// [skipPacketCount] is the number of packet to be skipped,
  /// For example, [skipPacketCount] is 2 by default.
  /// The plotting stream only update a half of packets
  /// which discards 1/2 packets.
  int skipPacketCount = 2;
}

extension BLEEcg on BLEManager {
  /// [startMeasurementEcg] sends read_ecg_packet to device,
  ///  and then, it registers all ecg streams of [streams].
  Future<void> startMeasurementEcg() async {
    await _sendCommandHolterTime();
    await _sendCommandReadEcg();

    if (!_isPaused) {
      streams
        ..rawStream = _getStream().broadcast
        ..decompressedStream = _getDecompressedStream().broadcast
        ..recentStream = _getRecentStream().broadcast
        ..leadStateStream = _getLeadStateStream().broadcast
        ..heartRateStream = _getHeartRateStream().broadcast
        ..bufferedStreamEcg = _getBufferedStream().broadcast
        ..bufferedFilterStreamEcg = _getBufferedFilterStream().broadcast
        ..bufferedPlotStreamEcg = _getBufferedPlotStream().broadcast
        ..bufferedPlotRTStreamEcg = _getBufferedPlotRTStream().broadcast;
    }

    _isMeasuring = true;
    _isPaused = false;
  }

  /// [pauseMeasurementEcg]
  Future<void> pauseMeasurementEcg() async {
    _isPaused = !_isPaused;
  }

  /// [stopMeasurementEcg] sends stop_ecg_packet to device.
  Future<void> stopMeasurementEcg() async {
    if (!_isPaused) _isMeasuring = false;
    final control = points[Characteristics.HEH_CONTROL_POINT];
    if (control == null) return;

    log('Writes a packet ${Commands.packetStopEcg}');
    await control.write(Commands.packetStopEcg);
  }

  Future<void> _sendCommandReadEcg() async {
    final control = points[Characteristics.HEH_CONTROL_POINT];
    if (control == null) return;

    log('Writes a packet ${Commands.packetReadEcg}');
    await control.write(Commands.packetReadEcg);
  }

  Future<void> _sendCommandHolterTime() async {
    final control = points[Characteristics.HEH_CONTROL_POINT];
    if (control == null) return;

    log('Writes a packet ${Commands.packetSetHolterTime}');
    await control.write(Commands.packetSetHolterTime);
  }

  Stream<List<int>> _getStream() async* {
    final measurement = points[Characteristics.HEH_MEASUREMENT];
    if (measurement == null) return;
    await measurement.setNotifyValue(true);
    await for (final res in measurement.value) {
      if (!_isMeasuring || device == null) break;
      if (_isPaused) continue;
      yield res;
    }
  }

  Stream<List<int>> _getDecompressedStream() async* {
    final events = streams.rawStream;
    if (events == null) return;
    await for (final event in events) {
      yield lzma.decode(event);
    }
  }

  Stream<List<int>>? get _events {
    return isCompressedEcg ? streams.decompressedStream : streams.rawStream;
  }

  Stream<List<double>> _getRecentStream() async* {
    final events = _events;
    if (events == null) return;
    final res = <double>[];
    await for (final event in events) {
      if (event.isEmpty) continue;
      final id = event.elementAtOrNull(3);
      if (id != 0x11) continue;
      final itemCount = event.length ~/ 3;
      res.clear();
      for (var i = 2; i < itemCount; i++) {
        final raw = toDouble(
          event[i * 3 + 0],
          event[i * 3 + 1],
          event[i * 3 + 2],
        );
        res.add(raw);
      }
      yield res;
    }
  }

  Stream<LeadState> _getLeadStateStream() async* {
    final events = _events;
    if (events == null) return;

    bool checkBit(int value, int bit) => (value & (1 << bit)) == 0;
    final leadState = LeadState.empty();

    var now = DateTime.now();

    await for (final event in events) {
      if (event.isEmpty) continue;
      final id = event.elementAtOrNull(3);
      if (id != 0x11) continue;

      final currentTime = DateTime.now();
      if (currentTime.difference(now) <= const Duration(milliseconds: 500)) {
        continue;
      }

      now = currentTime;

      final leadStateHigh = event.elementAtOrNull(4);
      final leadStateLow = event.elementAtOrNull(5);

      if (leadStateHigh == null || leadStateLow == null) continue;

      yield leadState.copyWith(
        ra: checkBit(leadStateHigh, 1),
        ll: checkBit(leadStateLow, 2),
        la: checkBit(leadStateLow, 1),
        v1: checkBit(leadStateLow, 7),
        v2: checkBit(leadStateLow, 3),
        v3: checkBit(leadStateLow, 4),
        v4: checkBit(leadStateLow, 5),
        v5: checkBit(leadStateLow, 6),
        v6: checkBit(leadStateLow, 0),
      );
    }
  }

  Stream<int> _getHeartRateStream() async* {
    final events = _events;
    if (events == null) return;

    await for (final event in events) {
      if (event.isEmpty) continue;
      final id = event.elementAtOrNull(3);
      if (id != 0x11) continue;

      final leadStateHigh = event.elementAtOrNull(4);
      final leadStateLow = event.elementAtOrNull(5);

      if (leadStateHigh != 0x00 || leadStateLow != 0x00) {
        yield 0;
        continue;
      }

      final itemCount = event.length ~/ 3;

      var count = 0;
      for (var i = 2; i < itemCount; i++) {
        count++;
        if (count % 8 != 2) continue;

        final raw = toDouble(
          event[i * 3 + 0],
          event[i * 3 + 1],
          event[i * 3 + 2],
        );

        final output = _adaptiveHR.calculateHR(
          raw,
          true,
        );

        if (output > 0) yield output;
      }
    }
  }

  Stream<Channels> _getBufferedStream() async* {
    final events = streams.recentStream;
    if (events == null) return;

    final res = List.generate(12, (index) => <double>[]);

    await for (final event in events) {
      for (var index = 0; index < 9; index++) {
        final l1 = event[index * 8 + 1];
        final l2 = event[index * 8 + 2];
        res[0].add(l1); // L1
        res[1].add(l2); // L2
        res[2].add(l2 - l1); // L3

        res[3].add(-(l1 + l2) / 2); // aVR
        res[4].add(l1 - l2 / 2); // aVL
        res[5].add(l2 - l1 / 2); // aVF

        res[6].add(event[index * 8 + 7]); // V1
        res[7].add(event[index * 8 + 3]); // V2
        res[8].add(event[index * 8 + 4]); // V3
        res[9].add(event[index * 8 + 5]); // V4
        res[10].add(event[index * 8 + 6]); // V5
        res[11].add(event[index * 8 + 0]); // V6
      }

      for (final channel in res) {
        if (channel.length > maxBufferStreamEcg) {
          channel.removeRange(0, channel.length - lengthBufferStreamEcg);
        }
      }
      yield Channels.fromList(res);
    }
  }

  Stream<Channels> _getBufferedFilterStream() async* {
    final events = streams.bufferedStreamEcg;
    if (events == null) return;

    await for (final event in events) {
      for (var i = event.l2.length - 9; i < event.l2.length; i++) {
        event.l1[i] = _dataFilter.applyFilter(0, event.l1[i], true);
        event.l2[i] = _dataFilter.applyFilter(1, event.l2[i], true);
        event.l3[i] = event.l2[i] - event.l1[i];

        event.aVR[i] = -(event.l2[i] + event.l1[i]) / 2;
        event.aVL[i] = event.l1[i] - event.l2[i] / 2;
        event.aVF[i] = event.l2[i] - event.l1[i] / 2;

        event.v1[i] = _dataFilter.applyFilter(2, event.v1[i], true);
        event.v2[i] = _dataFilter.applyFilter(3, event.v2[i], true);
        event.v3[i] = _dataFilter.applyFilter(4, event.v3[i], true);
        event.v4[i] = _dataFilter.applyFilter(5, event.v4[i], true);
        event.v5[i] = _dataFilter.applyFilter(6, event.v5[i], true);
        event.v6[i] = _dataFilter.applyFilter(7, event.v6[i], true);
      }
      yield event;
    }
  }

  Stream<PlotChannels> _getBufferedPlotStream() async* {
    // ignore: lines_longer_than_80_chars
    final events = isFilteredEcg ? streams.bufferedStreamEcg : streams.bufferedFilterStreamEcg;

    if (events == null) return;

    final resList = List.generate(12, (index) => <FlSpot>[]);
    final res = PlotChannels.fromList(resList, 0);
    var x = 0.0;

    await for (final event in events) {
      for (var i = event.l2.length - 9; i < event.l2.length; i++) {
        final j = i - event.l2.length + 9;

        res.l1.add(FlSpot(x + j, event.l1[i]));
        res.l2.add(FlSpot(x + j, event.l2[i]));
        res.l3.add(FlSpot(x + j, event.l3[i]));

        res.aVR.add(FlSpot(x + j, event.aVR[i]));
        res.aVL.add(FlSpot(x + j, event.aVL[i]));
        res.aVF.add(FlSpot(x + j, event.aVF[i]));

        res.v1.add(FlSpot(x + j, event.v1[i]));
        res.v2.add(FlSpot(x + j, event.v2[i]));
        res.v3.add(FlSpot(x + j, event.v3[i]));
        res.v4.add(FlSpot(x + j, event.v4[i]));
        res.v5.add(FlSpot(x + j, event.v5[i]));
        res.v6.add(FlSpot(x + j, event.v6[i]));
      }

      if (res.l2.length > event.l2.length) {
        res.l1.removeRange(0, res.l1.length - event.l1.length);
        res.l2.removeRange(0, res.l2.length - event.l2.length);
        res.l3.removeRange(0, res.l3.length - event.l3.length);

        res.aVR.removeRange(0, res.aVR.length - event.aVR.length);
        res.aVL.removeRange(0, res.aVL.length - event.aVL.length);
        res.aVF.removeRange(0, res.aVF.length - event.aVF.length);

        res.v1.removeRange(0, res.v1.length - event.v1.length);
        res.v2.removeRange(0, res.v2.length - event.v2.length);
        res.v3.removeRange(0, res.v3.length - event.v3.length);
        res.v4.removeRange(0, res.v4.length - event.v4.length);
        res.v5.removeRange(0, res.v5.length - event.v5.length);
        res.v6.removeRange(0, res.v6.length - event.v6.length);
      }

      yield res;
      x += 9;
    }
  }

  Stream<PlotChannels> _getBufferedPlotRTStream() async* {
    final events = streams.bufferedPlotStreamEcg;

    if (events == null) return;

    double skip(double x) {
      final newX = x % screenBufferStreamingEcg;
      final newXSkip = newX - newX % skipSampleCount;
      return newXSkip;
    }

    FlSpot f(FlSpot original) {
      return FlSpot(
        skip(original.x),
        original.y,
      );
    }

    final resList = List.generate(
      12,
      (_) => List<FlSpot>.generate(
        screenBufferStreamingEcg ~/ skipSampleCount,
        (index) => FlSpot.nullSpot,
        growable: false,
      ),
    );

    final res = PlotChannels.fromList(resList, 0);
    var x = 0;

    await for (final event in events) {
      for (var i = event.l2.length - 9; i < event.l2.length; i++) {
        final j = i - event.l2.length + 9;

        if (skipSampleCount != 1) {
          if ((x + j) % skipSampleCount != 0) continue;
        }

        final index = skip(x + j.toDouble());

        res.l1[index ~/ skipSampleCount] = f(event.l1[i]);
        res.l2[index ~/ skipSampleCount] = f(event.l2[i]);
        res.l3[index ~/ skipSampleCount] = f(event.l3[i]);

        res.aVR[index ~/ skipSampleCount] = f(event.aVR[i]);
        res.aVL[index ~/ skipSampleCount] = f(event.aVL[i]);
        res.aVF[index ~/ skipSampleCount] = f(event.aVF[i]);

        res.v1[index ~/ skipSampleCount] = f(event.v1[i]);
        res.v2[index ~/ skipSampleCount] = f(event.v2[i]);
        res.v3[index ~/ skipSampleCount] = f(event.v3[i]);
        res.v4[index ~/ skipSampleCount] = f(event.v4[i]);
        res.v5[index ~/ skipSampleCount] = f(event.v5[i]);
        res.v6[index ~/ skipSampleCount] = f(event.v6[i]);

        if (refreshBufferEcg > 0) {
          final refreshedIndex = skip(x + j.toDouble() + refreshBufferEcg);

          res.l1[refreshedIndex ~/ skipSampleCount] = FlSpot.nullSpot;
          res.l2[refreshedIndex ~/ skipSampleCount] = FlSpot.nullSpot;
          res.l3[refreshedIndex ~/ skipSampleCount] = FlSpot.nullSpot;

          res.aVR[refreshedIndex ~/ skipSampleCount] = FlSpot.nullSpot;
          res.aVL[refreshedIndex ~/ skipSampleCount] = FlSpot.nullSpot;
          res.aVF[refreshedIndex ~/ skipSampleCount] = FlSpot.nullSpot;

          res.v1[refreshedIndex ~/ skipSampleCount] = FlSpot.nullSpot;
          res.v2[refreshedIndex ~/ skipSampleCount] = FlSpot.nullSpot;
          res.v3[refreshedIndex ~/ skipSampleCount] = FlSpot.nullSpot;
          res.v4[refreshedIndex ~/ skipSampleCount] = FlSpot.nullSpot;
          res.v5[refreshedIndex ~/ skipSampleCount] = FlSpot.nullSpot;
          res.v6[refreshedIndex ~/ skipSampleCount] = FlSpot.nullSpot;
        }
      }

      x += 9;

      if (skipPacketCount == 1 || x ~/ 9 % skipPacketCount == 0) {
        yield res.copyWith(
          pivot: (x % screenBufferStreamingEcg).toDouble(),
        );
      }
    }
  }
}

==> healthrian_ble_support_for_wearecg12-1.0.19.tar.gz: lib/src/mixin/ecg.dart <==
// ignore_for_file: public_member_api_docs

import 'dart:developer';

import 'package:collection/collection.dart';
import 'package:fl_chart/fl_chart.dart';
import 'package:healthrian_ble_support_for_wearecg12/'
    'healthrian_ble_support_for_wearecg12.dart';
import 'package:healthrian_common_support/healthrian_common_support.dart';
import 'package:lzma/lzma.dart';

mixin Ecg {
  /// [maxBufferStreamEcg] decides a maximum length of buffered Stream.
  ///
  /// For 40 seconds buffer and 250 sampling rate,
  /// It uses 40 * 250 by default.
  int maxBufferStreamEcg = 40 * 250;

  /// [lengthBufferStreamEcg] decides a length of buffered Stream,
  /// after flushing buffer when reaching a maximum size.
  ///
  /// For 30 seconds buffer and 250 sampling rate,
  /// It uses 30 * 250 by default.
  int lengthBufferStreamEcg = 30 * 250;

  /// [screenBufferStreamingEcg] decides a length of buffered Stream,
  /// which is shown to screen.
  ///
  /// For 5 seconds buffer and 250 sampling rate,
  /// It uses 5 * 250 by default.
  int screenBufferStreamingEcg = 5 * 250;

  /// [refreshBufferEcg] decides a length to be hidden after refreshed.
  ///
  /// For 0.5 seconds range to be hidden and 250 sampling rate,
  /// It use 250 ~/ 2 by default.
  int refreshBufferEcg = 250 ~/ 2;

  /// [isFilteredEcg] decides whether ecg signal is filtered.
  bool isFilteredEcg = true;

  /// [isCompressedEcg] decides whether ecg signal is compressed.
  ///
  /// If it is `true`, ecg signal is decompressed by LZMA algorithm.
  bool isCompressedEcg = false;

  final _dataFilter = DataFilter(8);

  final _adaptiveHR = AdaptiveHR(0);

  bool _isPaused = false;

  bool _isMeasuring = false;

  /// [skipSample] is the number of samples to be skipped.
  /// For example, [skipSample] is 5 by default.
  /// The plotting stream only emits a 1/5 samples,
  /// which discards 4/5 samples.
  int skipSample = 5;

  /// [skipPacket] is the number of packet to be skipped,
  /// For example, [skipPacket] is 2 by default.
  /// The plotting stream only update a half of packets
  /// which discards 1/2 packets.
  int skipPacket = 2;

  final _channelCount = 12;
  final _sampleCount = 9;

  int get _totalCount => _channelCount * _sampleCount;
}

extension BLEEcg on BLEManager {
  /// [startMeasurementEcg] sends read_ecg_packet to device,
  ///  and then, it registers all ecg streams of [streams].
  Future<void> startMeasurementEcg() async {
    await _sendCommandHolterTime();
    await _sendCommandReadEcg();

    if (!_isPaused) {
      streams
        ..rawStream = _getStream().broadcast
        ..decompressedStream = _getDecompressedStream().broadcast
        ..recentStream = _getRecentStream().broadcast
        ..leadStateStream = _getLeadStateStream().broadcast
        ..heartRateStream = _getHeartRateStream().broadcast
        ..bufferedStreamEcg = _getBufferedStream().broadcast
        ..bufferedFilterStreamEcg = _getBufferedFilterStream().broadcast
        ..bufferedPlotStreamEcg = _getBufferedPlotStream().broadcast
        ..bufferedPlotRTStreamEcg = _getBufferedPlotRTStream().broadcast;
    }

    _isMeasuring = true;
    _isPaused = false;
  }

  /// [pauseMeasurementEcg]
  Future<void> pauseMeasurementEcg() async {
    _isPaused = !_isPaused;
  }

  /// [stopMeasurementEcg] sends stop_ecg_packet to device.
  Future<void> stopMeasurementEcg() async {
    _isMeasuring = false;
    final control = points[Characteristics.HEH_CONTROL_POINT];
    if (control == null) return;

    log('Writes a packet ${Commands.packetStopEcg}');
    await control.write(Commands.packetStopEcg);
  }

  Future<void> _sendCommandReadEcg() async {
    final control = points[Characteristics.HEH_CONTROL_POINT];
    if (control == null) return;

    log('Writes a packet ${Commands.packetReadEcg}');
    await control.write(Commands.packetReadEcg);
  }

  Future<void> _sendCommandHolterTime() async {
    final control = points[Characteristics.HEH_CONTROL_POINT];
    if (control == null) return;

    log('Writes a packet ${Commands.packetSetHolterTime}');
    await control.write(Commands.packetSetHolterTime);
  }

  Stream<List<int>> _getStream() async* {
    final measurement = points[Characteristics.HEH_MEASUREMENT];
    if (measurement == null) return;
    await measurement.setNotifyValue(true);
    await for (final res in measurement.value) {
      if (!_isMeasuring || device == null) break;
      if (_isPaused) continue;
      yield res;
    }
  }

  Stream<List<int>> _getDecompressedStream() async* {
    final events = streams.rawStream;
    if (events == null) return;
    await for (final event in events) {
      yield lzma.decode(event);
    }
  }

  Stream<List<int>>? get _events {
    return isCompressedEcg ? streams.decompressedStream : streams.rawStream;
  }

  Stream<List<double>> _getRecentStream() async* {
    final events = _events;
    if (events == null) return;
    final res = <double>[];
    await for (final event in events) {
      if (event.isEmpty) continue;
      final id = event.elementAtOrNull(3);
      if (id != 0x11) continue;
      final itemCount = event.length ~/ 3;
      res.clear();
      for (var i = 2; i < itemCount; i++) {
        final raw = toDouble(
          event[i * 3 + 0],
          event[i * 3 + 1],
          event[i * 3 + 2],
        );
        res.add(raw);
      }
      yield res;
    }
  }

  Stream<LeadState> _getLeadStateStream() async* {
    final events = _events;
    if (events == null) return;

    bool checkBit(int value, int bit) => (value & (1 << bit)) == 0;
    final leadState = LeadState.empty();

    var now = DateTime.now();

    await for (final event in events) {
      if (event.isEmpty) continue;
      final id = event.elementAtOrNull(3);
      if (id != 0x11) continue;

      final currentTime = DateTime.now();
      if (currentTime.difference(now) <= const Duration(milliseconds: 500)) {
        continue;
      }

      now = currentTime;

      final leadStateHigh = event.elementAtOrNull(4);
      final leadStateLow = event.elementAtOrNull(5);

      if (leadStateHigh == null || leadStateLow == null) continue;

      yield leadState.copyWith(
        ra: checkBit(leadStateHigh, 1),
        ll: checkBit(leadStateLow, 2),
        la: checkBit(leadStateLow, 1),
        v1: checkBit(leadStateLow, 7),
        v2: checkBit(leadStateLow, 3),
        v3: checkBit(leadStateLow, 4),
        v4: checkBit(leadStateLow, 5),
        v5: checkBit(leadStateLow, 6),
        v6: checkBit(leadStateLow, 0),
      );
    }
  }

  Stream<int> _getHeartRateStream() async* {
    final events = _events;
    if (events == null) return;

    await for (final event in events) {
      if (event.isEmpty) continue;
      final id = event.elementAtOrNull(3);
      if (id != 0x11) continue;

      final leadStateHigh = event.elementAtOrNull(4);
      final leadStateLow = event.elementAtOrNull(5);

      if (leadStateHigh != 0x00 || leadStateLow != 0x00) {
        yield 0;
        continue;
      }

      final itemCount = event.length ~/ 3;

      for (var i = 2; i < itemCount; i++) {
        if (i % 8 != 0) continue;

        final raw = toDouble(
          event[i * 3 + 0],
          event[i * 3 + 1],
          event[i * 3 + 2],
        );

        final output = _adaptiveHR.calculateHR(
          raw,
          true,
        );

        if (output > 0) yield output;
      }
    }
  }

  Stream<Channels> _getBufferedStream() async* {
    final events = streams.recentStream;
    if (events == null) return;

    final res = List.generate(_channelCount, (index) => <double>[]);

    await for (final event in events) {
      for (var index = 0; index < _sampleCount; index++) {
        final l1 = event[index * 8 + 1];
        final l2 = event[index * 8 + 2];
        res[0].add(l1); // L1
        res[1].add(l2); // L2
        res[2].add(l2 - l1); // L3

        res[3].add(-(l1 + l2) / 2); // aVR
        res[4].add(l1 - l2 / 2); // aVL
        res[5].add(l2 - l1 / 2); // aVF

        res[6].add(event[index * 8 + 7]); // V1
        res[7].add(event[index * 8 + 3]); // V2
        res[8].add(event[index * 8 + 4]); // V3
        res[9].add(event[index * 8 + 5]); // V4
        res[10].add(event[index * 8 + 6]); // V5
        res[11].add(event[index * 8 + 0]); // V6
      }

      for (final channel in res) {
        if (channel.length > maxBufferStreamEcg) {
          channel.removeRange(0, channel.length - lengthBufferStreamEcg);
        }
      }
      yield Channels.fromList(res);
    }
  }

  Stream<Channels> _getBufferedFilterStream() async* {
    final events = streams.bufferedStreamEcg;
    if (events == null) return;

    await for (final event in events) {
      for (var i = event.l2.length - _sampleCount; i < event.l2.length; i++) {
        event.l1[i] = _dataFilter.applyFilter(0, event.l1[i], true);
        event.l2[i] = _dataFilter.applyFilter(1, event.l2[i], true);
        event.l3[i] = event.l2[i] - event.l1[i];

        event.aVR[i] = -(event.l2[i] + event.l1[i]) / 2;
        event.aVL[i] = event.l1[i] - event.l2[i] / 2;
        event.aVF[i] = event.l2[i] - event.l1[i] / 2;

        event.v1[i] = _dataFilter.applyFilter(2, event.v1[i], true);
        event.v2[i] = _dataFilter.applyFilter(3, event.v2[i], true);
        event.v3[i] = _dataFilter.applyFilter(4, event.v3[i], true);
        event.v4[i] = _dataFilter.applyFilter(5, event.v4[i], true);
        event.v5[i] = _dataFilter.applyFilter(6, event.v5[i], true);
        event.v6[i] = _dataFilter.applyFilter(7, event.v6[i], true);
      }
      yield event;
    }
  }

  Stream<PlotChannels> _getBufferedPlotStream() async* {
    // ignore: lines_longer_than_80_chars
    final events = isFilteredEcg ? streams.bufferedStreamEcg : streams.bufferedFilterStreamEcg;

    if (events == null) return;

    final resList = List.generate(_channelCount, (index) => <FlSpot>[]);
    final res = PlotChannels.fromList(resList, 0);
    var x = 0.0;

    await for (final event in events) {
      for (var i = event.l2.length - _sampleCount; i < event.l2.length; i++) {
        final j = i - event.l2.length + _sampleCount;

        res.l1.add(FlSpot(x + j, event.l1[i]));
        res.l2.add(FlSpot(x + j, event.l2[i]));
        res.l3.add(FlSpot(x + j, event.l3[i]));

        res.aVR.add(FlSpot(x + j, event.aVR[i]));
        res.aVL.add(FlSpot(x + j, event.aVL[i]));
        res.aVF.add(FlSpot(x + j, event.aVF[i]));

        res.v1.add(FlSpot(x + j, event.v1[i]));
        res.v2.add(FlSpot(x + j, event.v2[i]));
        res.v3.add(FlSpot(x + j, event.v3[i]));
        res.v4.add(FlSpot(x + j, event.v4[i]));
        res.v5.add(FlSpot(x + j, event.v5[i]));
        res.v6.add(FlSpot(x + j, event.v6[i]));
      }

      if (res.l2.length > event.l2.length) {
        res.l1.removeRange(0, res.l1.length - event.l1.length);
        res.l2.removeRange(0, res.l2.length - event.l2.length);
        res.l3.removeRange(0, res.l3.length - event.l3.length);

        res.aVR.removeRange(0, res.aVR.length - event.aVR.length);
        res.aVL.removeRange(0, res.aVL.length - event.aVL.length);
        res.aVF.removeRange(0, res.aVF.length - event.aVF.length);

        res.v1.removeRange(0, res.v1.length - event.v1.length);
        res.v2.removeRange(0, res.v2.length - event.v2.length);
        res.v3.removeRange(0, res.v3.length - event.v3.length);
        res.v4.removeRange(0, res.v4.length - event.v4.length);
        res.v5.removeRange(0, res.v5.length - event.v5.length);
        res.v6.removeRange(0, res.v6.length - event.v6.length);
      }

      yield res;
      x += _sampleCount;
    }
  }

  Stream<PlotChannels> _getBufferedPlotRTStream() async* {
    final events = streams.bufferedPlotStreamEcg;

    if (events == null) return;

    double skip(double x) {
      final newX = x % screenBufferStreamingEcg;
      final newXSkip = newX - newX % skipSample;
      return newXSkip;
    }

    FlSpot f(FlSpot original) {
      return FlSpot(
        skip(original.x),
        original.y,
      );
    }

    final resList = List.generate(
      _channelCount,
      (_) => List<FlSpot>.generate(
        screenBufferStreamingEcg ~/ skipSample,
        (index) => FlSpot.nullSpot,
        growable: false,
      ),
    );

    final res = PlotChannels.fromList(resList, 0);
    var x = 0;

    await for (final event in events) {
      for (var i = event.l2.length - _sampleCount; i < event.l2.length; i++) {
        final j = i - event.l2.length + _sampleCount;

        if (skipSample != 1) {
          if ((x + j) % skipSample != 0) continue;
        }

        final index = skip(x + j.toDouble());

        res.l1[index ~/ skipSample] = f(event.l1[i]);
        res.l2[index ~/ skipSample] = f(event.l2[i]);
        res.l3[index ~/ skipSample] = f(event.l3[i]);

        res.aVR[index ~/ skipSample] = f(event.aVR[i]);
        res.aVL[index ~/ skipSample] = f(event.aVL[i]);
        res.aVF[index ~/ skipSample] = f(event.aVF[i]);

        res.v1[index ~/ skipSample] = f(event.v1[i]);
        res.v2[index ~/ skipSample] = f(event.v2[i]);
        res.v3[index ~/ skipSample] = f(event.v3[i]);
        res.v4[index ~/ skipSample] = f(event.v4[i]);
        res.v5[index ~/ skipSample] = f(event.v5[i]);
        res.v6[index ~/ skipSample] = f(event.v6[i]);

        if (refreshBufferEcg > 0) {
          final refreshedIndex = skip(x + j.toDouble() + refreshBufferEcg);

          res.l1[refreshedIndex ~/ skipSample] = FlSpot.nullSpot;
          res.l2[refreshedIndex ~/ skipSample] = FlSpot.nullSpot;
          res.l3[refreshedIndex ~/ skipSample] = FlSpot.nullSpot;

          res.aVR[refreshedIndex ~/ skipSample] = FlSpot.nullSpot;
          res.aVL[refreshedIndex ~/ skipSample] = FlSpot.nullSpot;
          res.aVF[refreshedIndex ~/ skipSample] = FlSpot.nullSpot;

          res.v1[refreshedIndex ~/ skipSample] = FlSpot.nullSpot;
          res.v2[refreshedIndex ~/ skipSample] = FlSpot.nullSpot;
          res.v3[refreshedIndex ~/ skipSample] = FlSpot.nullSpot;
          res.v4[refreshedIndex ~/ skipSample] = FlSpot.nullSpot;
          res.v5[refreshedIndex ~/ skipSample] = FlSpot.nullSpot;
          res.v6[refreshedIndex ~/ skipSample] = FlSpot.nullSpot;
        }
      }

      x += _sampleCount;

      if (skipPacket == 1 || x ~/ _sampleCount % skipPacket == 0) {
        yield res.copyWith(
          pivot: (x % screenBufferStreamingEcg).toDouble(),
        );
      }
    }
  }
}
